/*******************************************************************************************/
/**************************  Linear Feedback Shift Register for Input sequence  ************/
/*******************************************************************************************/
int lfsr()
{
int b,l2;

b = ( (l & (1 << 31) ) >> 31 ) ; 

l2 =   (l & 1) ^ ( (l & (1 << 1)) >> 1 ) ^ ( (l & (1 << 21)) >> 21 ) ^ b ;

l = (l << 1) | l2;

return(b);
}

/*************************************************************************/
/****** unif: generate a uniform distributed random variable *************/
/*************************************************************************/
double unif(void)
{
long s;
long m=0x7FFFFFFF,a=16807,q=127773,r=2836;

s=vc_un/q;
vc_un=a*(vc_un-q*s)-r*s;
if (vc_un<=0) vc_un+=m;

return ( (double) vc_un/m );
}

 

/***************************************************************************/
/******* Gaussian: Generate a Gaussian variable with variance var (global) */
/***************************************************************************/
double Gaussian()
{
static double f1,f2,f3,a;
static double x1,x2;
static int isready=0;
if(isready)
{
	isready=0;
	return x2;
}
else
{

	f3 = 100.0;
	while (f3 > 1.0)
	{
		f1 = 2.0 * unif() - 1.0 ;
		f2 = 2.0 * unif() - 1.0 ;
		f3 = f1*f1 + f2*f2;
	}
	a = var * sqrt(-2.0 * log(f3)/f3);  

	x2 = f2 * a;
	isready=1;
	return (f1 * a);
}
}



/*************************************************************************/
/********************   CHANNEL_FADING: Phase 2 Channel ******************/
/*************************************************************************/
void channel_fading()
{
int nnew,nold;
register int curr_slot,col,k,q,slot_dec,slot_ttt;
double x;
const double factor = (double)1.12201845; // Increase factor
nnew=1;
nold=0;

/* Modulation and interleaving */

/* Pilot symbols */
for(col=0;col< 4;col++)
{
	for (curr_slot=0;curr_slot<SLOT_PER_FRAME;curr_slot++)
	{
		 slot[curr_slot][col].Re=1.;
		 slot[curr_slot][col].Im=1.;
	}
}

/* Codeword */
k=0;
for(col=4;col< 4+ncod/(2*SLOT_PER_FRAME);col++)
{
	for (curr_slot=0;curr_slot<SLOT_PER_FRAME;curr_slot++)
	{
	  slot[curr_slot][col].Re=(double)(2.*codeword[k++]-1);
	}

	for (curr_slot=0;curr_slot<SLOT_PER_FRAME;curr_slot++)
	{
	  slot[curr_slot][col].Im=(double)(2.*codeword[k++]-1);
	}
}
curr_slot=0;

/* Tail bits */
while(k<ncod)
{
  slot[curr_slot][col].Re=(double)(2.*codeword[k++]-1);
  if(k<ncod)slot[curr_slot][col].Im=(double)(2.*codeword[k++]-1);
  else slot[curr_slot][col].Im=1.;
	curr_slot++;
}

  /* Channel  */
	for(curr_slot=0;curr_slot<SLOT_PER_FRAME;curr_slot++)
	{
		for( col = 0; col < symbols_per_slot; ++col ) 
		{
		/* Gaussian noise */
			noise_signal[col].Re1 = (double)Gaussian();
			noise_signal[col].Im1 = (double)Gaussian();
			noise_signal[col].Re2 = (double)Gaussian();
			noise_signal[col].Im2 = (double)Gaussian();
			noise_signal[col].Re3 = (double)Gaussian();
			noise_signal[col].Im3 = (double)Gaussian();
			noise_signal[col].Re4 = (double)Gaussian();
			noise_signal[col].Im4 = (double)Gaussian();
			noise_signal[col].Re5 = (double)Gaussian();
			noise_signal[col].Im5 = (double)Gaussian();
			noise_signal[col].Re6 = (double)Gaussian();
			noise_signal[col].Im6 = (double)Gaussian();
			noise_signal[col].Re7 = (double)Gaussian();
			noise_signal[col].Im7 = (double)Gaussian();

	/* Tap for the fading channel */
			ch_response[col]      = fade_lee(count++);

	/* Power control */

			slot[curr_slot][col].Re*=(double)level;
			slot[curr_slot][col].Im*=(double)level;
		}
	/* Slot that is currenty decoded (delay 3)*/
		slot_dec=(SLOT_PER_FRAME+curr_slot-3)%SLOT_PER_FRAME;

	/* Basic Demod */
		Basic_Demod(
					slot[curr_slot],
					&symbols_per_slot,
					noise_signal,
					&Rake_fingers,
					ch_response,
       			soft_output[slot_dec],
          		&pc_flag,
					&ave_signal_pwr_per_slot[slot_dec], 
					&ave_noise_pwr_per_slot[slot_dec]);

		/* Power measures  */
		if(f!=0 || curr_slot>2)
		{
			total_rx_pwr += ave_signal_pwr_per_slot[slot_dec];
			total_rx_nse += ave_noise_pwr_per_slot[slot_dec];
		}
		/* Power control */
		if      (pc_flag== POWER_UP)level*=factor;
		else                        level/=factor;


		/* One frame is ready for decoding */
		if(curr_slot==2)
		{	
			/* Channel Deinterleaver LLR computation and quantization */
			k=0;
			for(col=4;col < 4+ncod/(2*SLOT_PER_FRAME); col++)
			{
				for (slot_ttt=0;slot_ttt<SLOT_PER_FRAME;slot_ttt++)
				{
					
				/* Lucent weighting factor */
//					x=2.*soft_output[slot_ttt][col].Re/var2;

					/* HNS weighting factor */
					x=2.*sqrt(ave_signal_pwr_per_slot[slot_ttt]*2.)
						      /ave_noise_pwr_per_slot[slot_ttt]
								*soft_output[slot_ttt][col].Re;

				/* Quantization over eight bits (3 bits to the left of the point) 
				*/   
					x=x*8.;
					q=(int)floor(x);
					if (q > 127)   q = 127;	if (q < -128)  q = -128;
					s_c[k++] = q;
				}
				for (slot_ttt=0;slot_ttt<SLOT_PER_FRAME;slot_ttt++)
				{		
//					x=2.*soft_output[slot_ttt][col].Im/var2;
					x=2.*sqrt(ave_signal_pwr_per_slot[slot_ttt]*2.)
						      /ave_noise_pwr_per_slot[slot_ttt]
								*soft_output[slot_ttt][col].Im;
					x=x*8.;
					q=(int)floor(x);
					if (q > 127)   q = 127;	if (q < -128)  q = -128;
					s_c[k++] = q;
				}
			}
/* Tail bits */
			slot_ttt=0;
			while(k<ncod)
			{
//				x=2.*soft_output[slot_ttt][col].Re/var2;
					x=2.*sqrt(ave_signal_pwr_per_slot[slot_ttt]*2.)
						      /ave_noise_pwr_per_slot[slot_ttt]
								*soft_output[slot_ttt][col].Re;
				x=x*8.;
				q=(int)floor(x);
				if (q > 127)   q = 127;	if (q < -128)  q = -128;
				s_c[k++] = q;
				if(k<ncod)
				{
//				x=2.*soft_output[slot_ttt][col].Im/var2;
				x=2.*sqrt(ave_signal_pwr_per_slot[slot_ttt]*2.)
					 /ave_noise_pwr_per_slot[slot_ttt]
					 *soft_output[slot_ttt][col].Im;
				x=x*8.;
				q=(int)floor(x);
				if (q > 127)   q = 127;	if (q < -128)  q = -128;
				s_c[k++] = q;
				}
				slot_ttt++;
			}
		}
	}
}

